﻿#define DEBUGFORM
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Collections;
using System.Windows.Forms;
using System.IO;
using PLDAAPINet;
using mPCIE_ADIO16_8F;
using System.Threading;
using System.Runtime.InteropServices;

namespace HMAPInet
{
    
    /// <summary>
    /// Class HM_API which exposes library Inrefaces
    /// </summary>
    class HM_API
    {
        public static int Configure_Value; 
        public static UInt32 Callback_Counter = 0;
        public static UInt32 Counter_Check = 0;
        public static UInt32 cPCIeCount = 0;
        public static UInt32 PCICount = 0, PciIndex = 0, Ret = 0;
        public static bool HM_Thread_Start_Stop_Status = false;
        public static readonly IntPtr INVALID_HANDLE_VALUE = new IntPtr(-1);
        public static UInt32[] PciAgentIndex = new UInt32[3];
        public static structure.SPCISlotDescription[] PCISlotDescription = new structure.SPCISlotDescription[Constants.NB_DEV];
        public static structure.SPCIConfigHeader0[] PCIConfigHeader0 = new structure.SPCIConfigHeader0[Constants.NB_DEV];
        public static structure.SPCIInterruptTransfer[] PCIInterruptTransfer0 = new structure.SPCIInterruptTransfer[Constants.NB_DEV];
        public static UInt32 False_Interrupt_Counter = 0;
        public static bool Error_Handler_Close = false;
        public static bool Error_Handler_Object_Signal = true;
        public static bool Log_Data_Completed = false;
        public static string[] Str = new string[(PLDAAPI.FIFO_THRESHOLD_VALUE() / Constants.BUFF_4BYTE_WORDS)];

        /// <summary>
        /// TheEnd method is called , when Failure occurs while detecting board during PCI scan 
        /// or Failure happens while resource allocation.
        /// </summary>
        public static void TheEnd()
        {
            for (UInt32 i = 0; i < cPCIeCount; i++)
            {
                PLDAAPI.PldaReleaseInterrupt(PciAgentIndex[i]);
                PLDAAPI.PldaClearResources(PciAgentIndex[i]);
            }
        }

        /// <summary>
        /// DrvEnd method is called, when PCIe Board not found.
        /// </summary>
        public static void DrvEnd()
        {
            MessageBox.Show("Clearing driver");
            PLDAAPI.PldaClearDriver();
        }

        /// <summary>
        /// Registering a Control Handler Function
        /// </summary>
        /// <param name="CtrlType">Holds Control Handler Signals Status</param>
        /// <returns>Returns Control Handler Status</returns>
        public static unsafe bool CtrlHandler(PLDAAPI.CtrlTypes CtrlType)
        {
            UInt32 Iteration, RegVal;
            switch (CtrlType)
            {
                case PLDAAPI.CtrlTypes.CTRL_C_EVENT:
                case PLDAAPI.CtrlTypes.CTRL_CLOSE_EVENT:
                case PLDAAPI.CtrlTypes.CTRL_BREAK_EVENT:
                    if (cPCIeCount > 0)
                    {
                        for (Iteration = 0; Iteration < cPCIeCount; Iteration++)
                        {
                            RegVal = (UInt32)DMA_CTRL.DMA_CTRL_ABORT;
                            PLDAAPI.PldaMemoryWrite32(Iteration, (UInt32)BAR.BAR_0, (UInt32)DMA.DMA_C2H_CTRL, 1, &RegVal);
                            PLDAAPI.PldaReleaseInterrupt(PciAgentIndex[Iteration]);
                            PLDAAPI.PldaClearResources(PciAgentIndex[Iteration]);

                        }
                        PLDAAPI.PldaClearDriver();
                    }
                    return true;

                default:
                    return false;
            }

        }

        /// <summary>
        /// Memory_Write Function is called to write data to the device memory
        /// </summary>
        /// <param name="DeviceIndex">Index of the device where writing should happen</param>
        /// <param name="Offset">Offset of the device where the data has to be written</param>
        /// <param name="Data">Data to be written</param>
        public static void Memory_Write(UInt32 DeviceIndex, UInt32 Offset, UInt32 Data)
        {
            UInt32 Count = 0;
            UInt32[] WriteBuffer = new UInt32[3];
            for (Count = 0; Count < Constants.BUFFER_SIZE_LENGTH; Count++)
            {
                WriteBuffer[Count] = Data;
            }
            PLDAAPI.PldaMemoryWrite32(DeviceIndex, RegisterOffset.BAR_REGISTER, (Offset / Constants.BUFF_4BYTE_WORDS), Constants.BUFFER_SIZE_LENGTH, WriteBuffer);
        }

        /// <summary>
        /// Memory_Read Function is called to read data from the device memory
        /// </summary>
        /// <param name="DeviceIndex">Index of the device where writing should happen</param>
        /// <param name="Offset">Offset of the device where the data has to be written</param>
        /// <returns>Returns the read data</returns>
        public static UInt32 Memory_Read(UInt32 DeviceIndex, UInt32 Offset)
        {
            UInt32[] ReadBuffer = new UInt32[3];
            UInt32 IndexOne = 1;
            PLDAAPI.PldaMemoryRead32(DeviceIndex, RegisterOffset.BAR_REGISTER, (Offset / Constants.BUFF_4BYTE_WORDS), Constants.BUFFER_SIZE_LENGTH, ref ReadBuffer[Constants.BUFFER_SIZE_LENGTH]);
            return ReadBuffer[IndexOne];
        }

        
        /// <summary>
        /// Open Function is used to open device
        /// </summary>
        /// <returns>Returns the index of opened device</returns>
        public static int Open()
        {

            
            if (PLDAAPI.Creating_ISR_Event() != true)
                MessageBox.Show("ERROR: Event Creation not happend properly");

            if (PLDAAPI.SetConsoleCtrlHandler(new PLDAAPI.HandlerRoutine(CtrlHandler), true) == false)
            {
                MessageBox.Show("ERROR: Could not set control handler");
                //return 0;
            }
            // Driver initialization for PLDA devices
            UInt32 IsDriverOK = PLDAAPI.PldaInitDriverFilter(Constants.PLDA_VENDOR_ID, 0);

            //  Check for Driver Version
            if (IsDriverOK == Convert.ToInt32(Status.DRIVER_FAILED_VERSION))
            {
                MessageBox.Show("Wrong Driver version", "Driver version");
             return Constants.ERROR;
            }
            else if (IsDriverOK == Convert.ToInt32(Status.DRIVER_FAILED_OPEN))
            {
                MessageBox.Show("Failed to initialise driver", "initialise driver");
              return Constants.ERROR;
            }
            // Get boards count
            UInt32 PCIAgentCount = 1;
            PLDAAPI.PldaGetPCIAgentCount(ref PCIAgentCount);
            if (PCIAgentCount == Constants.ZERO_BOARDS)
            {
                MessageBox.Show(Convert.ToString(PCIAgentCount), " Failed Number of PCI boards");
                DrvEnd();
            }
            // Searching for PLDA cards in all PCIe ports
            cPCIeCount = 0;
            
            for (PCICount = 0; PCICount < PCIAgentCount; PCICount++)
            {
                //Retrieving informations of PCI agent number
                if (PLDAAPI.PldaGetPCIAgentID(PCICount, ref PCISlotDescription[cPCIeCount]) != 1)
                {
                    MessageBox.Show("Error retrieving informations of PCI agent number");
                    continue;
                }

                if ((PCISlotDescription[cPCIeCount].VendorID) == Constants.PLDA_VENDOR_ID)
                {
                    PciAgentIndex[cPCIeCount] = PCICount;
                    cPCIeCount++;
                }
            }
            if (cPCIeCount == Constants.ZERO_BOARDS)
            {
                MessageBox.Show("PLDA PCIe Board not found");
                DrvEnd();
            }
            //	Scan and report all detected PLDA boards
            for (PciIndex = 0; PciIndex < cPCIeCount; PciIndex++)
            {
                // Lock resources
                if ((Ret = PLDAAPI.PldaLockResources(PciAgentIndex[PciIndex])) != 1)
                    MessageBox.Show(string.Format("Board {0} resources lock : Failed ", PciAgentIndex[PciIndex]));

                if (PLDAAPI.PldaGetConfigHeader(PciAgentIndex[PciIndex], ref PCIConfigHeader0[PciIndex]) == 0)
                    MessageBox.Show(string.Format("Board {0} configuration : Failed ", PciAgentIndex[PciIndex]));
                else
                {
                    if (PCIConfigHeader0[PciIndex].BaseClass != Constants.EZDMA_BASE_CLASS)
                        MessageBox.Show("UNKNOWN DESIGN !!!");

                }
                
                PLDAAPI.PldaClearResources(PciAgentIndex[PciIndex]);

            }

            // If more than 1 board is detected
            if (cPCIeCount > 1)
                MessageBox.Show("More than one board is detected");
            // Searching for PLDA cards in all PCIe ports. Check for PLDA cards in all PCIe ports
            // This procedure should be executed once more to enumerate and order the boards


            cPCIeCount = 0;
            for (PCICount = 0; PCICount < PCIAgentCount; PCICount++)
            {

                if (PLDAAPI.PldaGetPCIAgentID(PCICount, ref PCISlotDescription[cPCIeCount]) == Constants.PCI_ERROR)
                {
                    MessageBox.Show("Error retrieving informations of PCI agent number");
                    continue;
                }

                if ((PCISlotDescription[cPCIeCount].VendorID) == Constants.PLDA_VENDOR_ID)
                {
                    PciAgentIndex[cPCIeCount] = PCICount;
                    cPCIeCount++;
                }
            }
            if (cPCIeCount == Constants.ZERO_BOARDS)
            {
                MessageBox.Show("PLDA PCIe Board not found");
                DrvEnd();
            }

            // Lock resources for the selected board
            if ((Ret = PLDAAPI.PldaLockResources(PciAgentIndex[PciIndex])) != 1)
            {
                MessageBox.Show(string.Format("Board {0} resources lock : Failed ", PciAgentIndex[PciIndex]));
                TheEnd();
            }

            // Initialize board interrupt
            if (PLDAAPI.PldaGetConfigHeader(PciAgentIndex[PciIndex], ref PCIConfigHeader0[PciIndex]) == Constants.PCI_ERROR)
            {
                MessageBox.Show(string.Format("Board {0} configuration : Failed ", PciAgentIndex[PciIndex]));
                TheEnd();
            }
            else
            {
                if (PCIConfigHeader0[PciIndex].BaseClass != Constants.EZDMA_BASE_CLASS)
                {
                    MessageBox.Show("Unknown design");
                    TheEnd();
                }
                else
                {
                    PciIndex = 0;
                    PCIInterruptTransfer0[PciIndex].BarIndex = (UInt32)BAR.BAR_2;
                    PCIInterruptTransfer0[PciIndex].AccessType = (UInt32)ISR.ISR_RW;
                    PCIInterruptTransfer0[PciIndex].ByteOffset = (UInt32)DMA.IRQ_CONTROL * sizeof(UInt32); //(UInt32)DMA.INT_REG * sizeof(UInt32);
                    PCIInterruptTransfer0[PciIndex].ResetValue = Constants.RESETVALUE;
                    PCIInterruptTransfer0[PciIndex].Ulmask = Constants.ULMASK;
                    PCIInterruptTransfer0[PciIndex].Uinbvec = Constants.UINBVEC;
                    PCIInterruptTransfer0[PciIndex].UlInterruptType = (UInt32)INTERRUPTTYPE.PIN_BASED_INT;
                    if (PLDAAPI.InitializeInterrupt(PciAgentIndex[PciIndex], ref PCIInterruptTransfer0[PciIndex]) != Constants.BOARD_INTERRUPT)
                    {
                        MessageBox.Show("Board interrupt init : FAILED");
                        TheEnd();
                    }
                }
            }
            StringBuilder GetAPIVer = new StringBuilder(50);
            PLDAAPI.PldaGetAPIVer(GetAPIVer);

            StringBuilder GetDrvVer = new StringBuilder(50);
            PLDAAPI.PldaGetDrvVer(GetDrvVer);

            //Enabling BITS(0-7) in  Reg 40 
            //Memory_Write(PciAgentIndex[PciIndex], RegisterOffset.IRQ_CONTROL_REG, Constants.IRQ_Enable);

            //DIO Configuration
            Memory_Write(PciAgentIndex[PciIndex], RegisterOffset.DIO_CONFIGURE, Constants.DIO_Configure);

            return (Convert.ToInt32(PciAgentIndex[PciIndex]));
        }



        /// <summary>
        /// When SGDMA mode selected , Thread_Start_Stop method will Holds the status of SGDMA mode.
        /// </summary>
        /// <param name="Status">Status of mode</param>
        public static void Thread_Start_Stop(bool Status)
        {
            HM_Thread_Start_Stop_Status = Status;
        }

        /// <summary>
        /// Abort the DMA operation
        /// </summary>
        public static unsafe void DMA_Abort(UInt32 DeviceIndex)
        {
            UInt32 RegVal = (UInt32)DMA_CTRL.DMA_CTRL_ABORT;
            PLDAAPI.PldaMemoryWrite32(DeviceIndex, (UInt32)BAR.BAR_0, (UInt32)DMA.DMA_C2H_CTRL, 1, &RegVal);
        }

        /// <summary>
        /// Configuring registers for data transfer in SGDMA Mode
        /// </summary>
        /// <param name="IBoard">Index of the device where writing should happen</param>
        /// <param name="ReadBuffer">Structure which holds Physical address, Bytecount, useraddress for SGDMA configuration</param>
        public static unsafe void Configure_SGDMA(UInt32 UBoard, structure.SSGBufferDescription ReadBuffer)
        {
            UInt32 RegVal = 0;
            RegVal = ReadBuffer.DescAddrLow;
            PLDAAPI.PldaMemoryWrite32(UBoard, (UInt32)BAR.BAR_0, (UInt32)DMA.DMA_C2H_LSB_ADDR, 1, &RegVal);

            RegVal = ReadBuffer.DescAddrHigh;
            PLDAAPI.PldaMemoryWrite32(UBoard, (UInt32)BAR.BAR_0, (UInt32)DMA.DMA_C2H_MSB_ADDR, 1, &RegVal);

            RegVal = ReadBuffer.ByteCount;
            PLDAAPI.PldaMemoryWrite32(UBoard, (UInt32)BAR.BAR_0, (UInt32)DMA.DMA_C2H_SIZE, 1, &RegVal);

            // start dma
            RegVal = (UInt32)DMA_CTRL.DMA_CTRL_START | (UInt32)DMA_CTRL.DMA_CTRL_SC_ENA;
            PLDAAPI.PldaMemoryWrite32(UBoard, (UInt32)BAR.BAR_0, (UInt32)DMA.DMA_C2H_CTRL, 1, &RegVal);
        }

        /// <summary>
        /// Log ADC data into a file depends on object signal
        /// </summary>
        /// <param name="IBoard">Index of the device where writing should happen</param>
        /// <param name="ReadBuffer">Structure which holds Physical address, Bytecount, useraddress for SGDMA configuration</param>
        public static unsafe void Interrupt_Check(UInt32 UBoard, structure.SSGBufferDescription ReadBuffer)
        {

            UInt32 Data = 0;
            bool Interrupt_Faliure_Status = false;
            
            while (Interrupt_Faliure_Status == false)
            {

                PLDAAPI.SetEvent(PLDAAPI.WakeThreadEventHandle(UBoard));
                UInt32 WaitResult = PLDAAPI.WaitForSingleObject(PLDAAPI.IRQ_EventHandle(UBoard), Constants.DMA_TIME_OUT_MS);

                if (WaitResult != Constants.WAIT_OBJECT_0)
                {
                    Interrupt_Faliure_Status = true;
                    Error_Handler_Object_Signal = false; // Convert.ToBoolean(Constants.WAIT_OBJECT_0);
                }

                else
                {
                    if (Counter_Check < PLDAAPI.IntStatus())
                    {
                        Counter_Check = PLDAAPI.IntStatus();
                        Callback_Counter++;
                        for (Data = 0; Data < ReadBuffer.ByteCount / Constants.BUFF_4BYTE_WORDS; Data++)
                        {
                            
                            Str[Data] = Convert.ToString(*((UInt32*)ReadBuffer.UserAddr + Data), 16);
                        }
                        if (Form1.Display_Flag == true)
                        {
                            //Periodically read SGDMA data 
                            Form1.Display_Flag = false;
                            Array.Copy(Str, Form1.Display_Str, HM_API.Str.Length);
                        }

                        Form1.Display_ADC_Value = true;

                        System.IO.File.AppendAllLines(AppDomain.CurrentDomain.BaseDirectory + @"\" + FileName.ADCLOG_FILE, Str);

                        Interrupt_Faliure_Status = true;
                    }
                    else
                    {
                        False_Interrupt_Counter++;
                        Interrupt_Faliure_Status = false;
                    }

                }

            }
        }

        /// <summary>
        ///  Main SGDMA function where Buffer Memory allocation for data transfer and Data logging into a file will happens.
        /// </summary>
        /// <param name="IBoard">Index of the device obtained during scan(PCI SCAN)</param>
        public static unsafe void SGDMA(object IBoard)
        {
            Form1.Display_Flag = true;
            Callback_Counter = 0;
            UInt32 UBoard = Convert.ToUInt32(IBoard);

            UInt32 Board = UBoard;
            structure.SSGBufferDescription Buffer = new structure.SSGBufferDescription();
            UInt32 Size = PLDAAPI.FIFO_THRESHOLD_VALUE();
            UInt32 Card_2_Host_Index;

            Buffer.ByteCount = Size;

            //Waiting for AD_Trigger_Interrupt 
            for (; ;)
            {
                if (Counter_Check < PLDAAPI.IntStatus())
                {
                    Counter_Check = PLDAAPI.IntStatus();
                    //File.AppendAllText("ADC_Trigger_Enable.txt", "AD_Trigger_IRQ : " + Convert.ToString(PLDAAPI.IntStatus()) + "\n");
                    //Enabling BITS(0-7) in  Reg 40 
                    HM_API.Memory_Write(UBoard, RegisterOffset.IRQ_CONTROL_REG, Constants.IRQ_Enable);
                    break;
                }
                if (HM_Thread_Start_Stop_Status == true)
                    break;
            }


            if (PLDAAPI.PldaSGMemLock(UBoard, &Card_2_Host_Index, &Buffer) == 0)
                ;//handle mem alloc error
            
            File.AppendAllText(FileName.ADCLOG_FILE, "********************************Logging ADC DATA********************************\n");
            File.AppendAllText(FileName.ADCLOG_FILE, " \n" + "Selected Frequency : " + Convert.ToString(Form1.FrequencyRate) + "\nSelected Transfer Size : " + Size + "\nSGDMA Enabled \n");
            File.AppendAllText(FileName.ADCLOG_FILE, "Data Logging Started : " + DateTime.Now.ToString("F") + "\n");
            PLDAAPI.MemSet((IntPtr)(UInt32*)Buffer.UserAddr, 0xAA, (int)Buffer.ByteCount);


            while (true)
            {

                PLDAAPI.MemSet((IntPtr)(UInt32*)Buffer.UserAddr, 0xAA, (int)Buffer.ByteCount);

                if (HM_Thread_Start_Stop_Status == true)
                {
                    Log_Data_Completed = true;
                    HM_Thread_Start_Stop_Status = false;
                    break;
                }

                Configure_SGDMA(UBoard, Buffer);

                Interrupt_Check(UBoard, Buffer);

                if (Error_Handler_Close == true && Error_Handler_Object_Signal == false)
                {
                    Error_Handler_Close = false;
                    Error_Handler_Object_Signal = true;
                    break;
                }

            }
            PLDAAPI.PldaSGMemUnlock(UBoard, Card_2_Host_Index);
            File.AppendAllText(FileName.ADCLOG_FILE, " \n" + "Test Completed on " + DateTime.Now.ToString("F") + "\nTotal count value : " + Callback_Counter + " \nTotal AD Samples Received : " + (Callback_Counter * Size/4) + "\n");
        }


        /// <summary>
        /// Holds and update the target frequency into offset 0x10 for ADC conversion
        /// </summary>
        /// <param name="DeviceIndex">Index of the device Obtained during scan</param>
        /// <param name="TargetHertz">Acquisition rate or Sampling frequency used for data transfer</param>
        public static void ADC_SetRate(UInt32 DeviceIndex, Double TargetHertz) //Changing frequency data to double for 0.1Hz option in GUI
        {
            UInt32 DeviceFrequency = Memory_Read(DeviceIndex, RegisterOffset.DEVICE_FREQUENCY_OFFSET);
            UInt32 Target_Frequency;
            Target_Frequency = PLDAAPI.Get_Target_Hz();
#if DEBUGFORM
            UInt32 RateDivisor = Convert.ToUInt32(DeviceFrequency / TargetHertz);//For 0.1Hz option in GUI
#else
            UInt32 RateDivisor = Convert.ToUInt32(DeviceFrequency / Target_Frequency);
#endif
            Memory_Write(DeviceIndex, RegisterOffset.ADC_RATE_DIVISOR_OFFSET, RateDivisor);
        }

        /// <summary>
        /// Update the  FIFO threshold value defined by user into a register
        /// </summary>
        /// <param name="DeviceIndex">Index of the device obtained during scan</param>
        /// <param name="Threshold">Transfer Size defined by user</param>
        public static void FIFO_Threshold(UInt32 DeviceIndex, UInt32 Threshold)
        {
            Memory_Write(DeviceIndex, RegisterOffset.FIFO_THRESHOLD_OFFSET, Threshold); // write into register 20
        }


       
        public static UInt32 ScanChannel = 7; //INx 0 0 0 (default)
        public static UInt32 Gain = 7;  // +_ 20.48 1 1 1 (default)
        public static UInt32 Sequence_Status = 0;
        public static UInt32[] Channel_Gain = new UInt32[8];
        public static UInt32 Channel_counter = 0;


        /// <summary>
        /// HM_Scan_channel method is used to hold and update the count of selected channels(channel range  0 - 7)
        /// </summary>
        /// <param name="ScanInput">Selected channel count</param>
        public static void HM_Scan_channel(UInt32 ScanInput) //Scan channel 0 through
        {
            ScanChannel = ScanInput;
        }

        /// <summary>
        /// HM_PGIA method is used to hold and update the gain of each channel
        /// </summary>
        /// <param name="PGIAValue"> Channel Gain</param>
        /// <param name="InputChannel">Selected channel</param>
        public static void HM_PGIA(UInt32 PGIAValue, UInt32 InputChannel)
        {
            Gain = PGIAValue;

        }

        public static string Compare_ADC_range_selection = "";
        public static bool Range_Set = false;
        /// <summary>
        /// HM_Sequencer_mode method is used to update sequencer modes(Basic and Advanced) depend on parameters 
        /// "Channel selection ", "Selection between Diffrential or Single ended", "Range selection"
        /// </summary>
        /// <param name="Range_Collector"> Collects Channel Ranges(Voltage)</param>
        /// <param name="COMValue"> Value based on basic sequencer or Advance sequencer mode.</param>
        public static void HM_Sequencer_mode(string[] Range_Collector, bool COMValue)
        {
            UInt32 Counter;
            if (COMValue == true) //Single ended mode
            {
                Compare_ADC_range_selection = Range_Collector[0];

                for (Counter = 0; Counter <= ScanChannel; Counter++)
                {
                    if (Compare_ADC_range_selection == Range_Collector[Counter])
                        Range_Set = false;
                    else
                    {
                        Range_Set = true;
                        break;
                    }

                }
                
                if (Range_Set == false)
                    Sequence_Status = Constants.BASIC_SEQUENCER;
                else
                    Sequence_Status = Constants.ADVANCE_SEQUENCER;


                for (UInt32 VAR_Gain = 0; VAR_Gain < Range_Collector.Length; VAR_Gain++) //Range_Collector.Length
                {

                    switch (Range_Collector[VAR_Gain])
                    {
                        case "±12":
                            Channel_Gain[VAR_Gain] = (UInt32)INX.IN0;
                            break;
                        case "±5":
                            Channel_Gain[VAR_Gain] = (UInt32)INX.IN1;
                            break;
                        case "±2.5":
                            Channel_Gain[VAR_Gain] = (UInt32)INX.IN2;
                            break;
                        case "±1.25":
                            Channel_Gain[VAR_Gain] = (UInt32)INX.IN3;
                            break;
                        case "±0.6125":
                            Channel_Gain[VAR_Gain] = (UInt32)INX.IN4;
                            break;
                        case "±0.3125":
                            Channel_Gain[VAR_Gain] = (UInt32)INX.IN5;
                            break;
                        case "reserved":
                            Channel_Gain[VAR_Gain] = (UInt32)INX.IN6;
                            break;
                        case "±10":
                            Channel_Gain[VAR_Gain] = (UInt32)INX.IN7;
                            break;
                        default:
                            break;

                    }
                }
            }
            else
            {
                Compare_ADC_range_selection = Range_Collector[0];

                for (Counter = 0; Counter <= ScanChannel; Counter += 2)
                {
                    if (Compare_ADC_range_selection == Range_Collector[Counter])
                        Range_Set = false;
                    else
                    {
                        Range_Set = true;
                        break;
                    }

                }
                
                if (Range_Set == false)
                    Sequence_Status = Constants.BASIC_SEQUENCER;
                else
                    Sequence_Status = Constants.ADVANCE_SEQUENCER;

                for (UInt32 VAR_Gain = 0; VAR_Gain < Range_Collector.Length; VAR_Gain += 2) //Range_Collector.Length
                {

                    switch (Range_Collector[VAR_Gain])
                    {
                        case "±12":
                            Channel_Gain[VAR_Gain] = (UInt32)INX.IN0;
                            break;
                        case "±5":
                            Channel_Gain[VAR_Gain] = (UInt32)INX.IN1;
                            break;
                        case "±2.5":
                            Channel_Gain[VAR_Gain] = (UInt32)INX.IN2;
                            break;
                        case "±1.25":
                            Channel_Gain[VAR_Gain] = (UInt32)INX.IN3;
                            break;
                        case "±0.6125":
                            Channel_Gain[VAR_Gain] = (UInt32)INX.IN4;
                            break;
                        case "±0.3125":
                            Channel_Gain[VAR_Gain] = (UInt32)INX.IN5;
                            break;
                        case "reserved":
                            Channel_Gain[VAR_Gain] = (UInt32)INX.IN6;
                            break;
                        case "±10":
                            Channel_Gain[VAR_Gain] = (UInt32)INX.IN7;
                            break;
                        default:
                            break;

                    }
                }
            }

        }

        /// <summary>
        /// Updating  Channel gain's and Configurations into Base+10 and Base+38 register
        /// </summary>
        /// <param name="DeviceIndex">Index of device</param>
        /// <param name="StartStop">Status of execution (Start or stop)</param>
        /// <param name="Configuration_Load">Holds Configuration_Load Status</param>
        /// <param name="COMValue"> Value based on basic sequencer or Advance sequencer mode.</param>
        public static void ADC_Continous_Control(UInt32 DeviceIndex, UInt32 StartStop, UInt32 Configuration_Load, bool COMValue)
        {
            UInt32 Reg = 0, Config_Reg = 0;
            if (COMValue == true) //Single ended mode
            {
                Reg = (Channel_Gain[3]) | (Channel_Gain[2] << ADC_GAIN_REGISTER.PGIA_GAIN_2) |
                                          (Channel_Gain[1] << ADC_GAIN_REGISTER.PGIA_GAIN_1) |
                                          (Channel_Gain[7] << ADC_GAIN_REGISTER.PGIA_GAIN_7) |
                                          (Channel_Gain[6] << ADC_GAIN_REGISTER.PGIA_GAIN_6) |
                                          (Channel_Gain[5] << ADC_GAIN_REGISTER.PGIA_GAIN_5) |
                                          (Channel_Gain[4] << ADC_GAIN_REGISTER.PGIA_GAIN_4) ;

                Memory_Write(DeviceIndex, RegisterOffset.ADC_GAIN, Reg);
            }

            else
            {
                Reg = (Channel_Gain[2] << ADC_GAIN_REGISTER.PGIA_GAIN_2) |
                      (Channel_Gain[6] << ADC_GAIN_REGISTER.PGIA_GAIN_6) |
                      (Channel_Gain[4] << ADC_GAIN_REGISTER.PGIA_GAIN_4) ;

                Memory_Write(DeviceIndex, RegisterOffset.ADC_GAIN, Reg);

            }


            Config_Reg =        (ADCContinuousControl.Conversion_Mode_Select    << ADC_CONFIGURATION.CMS)              |  
                                (ADCContinuousControl.Temperature_Sensor_Enable << ADC_CONFIGURATION.TEMPB)            | 
                                (Sequence_Status                                << ADC_CONFIGURATION.SEQ)              | 
                                (ADCContinuousControl.Multiplexer               << ADC_CONFIGURATION.MUX )             |
                                (Channel_Gain[0]                                << ADC_CONFIGURATION.PGIA_GAIN_0)      | 
                                 Convert.ToUInt32(COMValue)                     << ADC_CONFIGURATION.COM               |
                                (ScanChannel                                    << ADC_CONFIGURATION.GAIN_CHANNEL)     | 
                                (StartStop                                      << ADC_CONFIGURATION.START_STOP)       |
                                (Configuration_Load                             << ADC_CONFIGURATION.CONFIG_LOAD)      ;

            Memory_Write(DeviceIndex, RegisterOffset.ADC_C0NTINUOUS_CONTROL_OFFSET, Config_Reg);

 
        }

        /// <summary>
        /// DIO_Read is called to read DIO registers
        /// </summary>
        /// <param name="DeviceIndex">Index of the device</param>
        /// <param name="Data">Read data which is then converted into byte format</param>
        public static void DIO_Read(UInt32 DeviceIndex, out byte[] Data)
        {
            UInt32[] ReadBuffer = new UInt32[1];
            UInt32 Byte = 0;

            //Reads DIO registers
            PLDAAPI.PldaMemoryRead32(DeviceIndex, RegisterOffset.BAR_REGISTER, RegisterOffset.DIO_DATA / Constants.BUFF_4BYTE_WORDS, Constants.BUFFER_SIZE_LENGTH, ref ReadBuffer[Byte]);

            Data = BitConverter.GetBytes(ReadBuffer[Byte]);

        }

        /// <summary>
        /// DIO_Byte_Configure function is used to configure ByteIndex 0 as 
        /// Input or Output based on user selection
        /// </summary>
        /// <param name="DeviceIndex">Index of the device</param>
        /// <param name="DIOByte0isOutput">DIOByte0isOutput is the bool value configuration selected 
        /// by the user and which has to be written 
        /// into DIO ByteIndex0 config register</param>
        public static unsafe void DIO_Byte_Configure(UInt32 DeviceIndex, bool DIOByte0isOutput)
        {
            UInt32[] WriteBuffer = new UInt32[1];

            UInt32 RegVal;

            //Setting Configure Register values based on user direction selection
            if (DIOByte0isOutput == false)
                Configure_Value = ~(1 << Constants.INPUT_MODE) & Configure_Value;
            else
                Configure_Value = Configure_Value | Constants.OUTPUT_MODE;

            RegVal = Convert.ToUInt32(Configure_Value);
            WriteBuffer[0] = RegVal;

            //Writes configure register values to the memory
            PLDAAPI.PldaMemoryWrite32(DeviceIndex, RegisterOffset.BAR_REGISTER, RegisterOffset.DIO_CONFIGURE / Constants.BUFF_4BYTE_WORDS, Constants.BUFFER_SIZE_LENGTH, WriteBuffer);
        }

        /// <summary>
        /// DIO_Bit_Configure function is used to configure I/O Group 1-8 bits as 
        /// Input or Output based on user selection
        /// </summary>
        /// <param name="DeviceIndex">Index of the device</param>
        /// <param name="DIOByte1Direction">DIOByte1Direction is the bool value configuration selected
        /// by the user and which has to be written 
        /// into DIO I/O Group 1-8 config register</param></param>
        public static void DIO_Bit_Configure(UInt32 DeviceIndex, bool[] DIOByte1Direction)
        {
            UInt32[] WriteBuffer = new UInt32[1];
            UInt32 RegVal;
            UInt32 Wb = 0;

            for (int Bit_Pos = 0; Bit_Pos < Constants.BIT_LENGTH; Bit_Pos++)
            {
                //Setting Configure Register values based on user direction selection
                if (DIOByte1Direction[Bit_Pos] == true)
                {
                    Configure_Value = (1 << (Bit_Pos + Constants.BYTEINDEX0_BITS)) | Configure_Value;
                }
                else
                {
                    Configure_Value = ~(1 << (Bit_Pos + Constants.BYTEINDEX0_BITS)) & Configure_Value;
                }
            }
            RegVal = Convert.ToUInt32(Configure_Value);
            WriteBuffer[Wb] = RegVal;
            //Writes configure register values to the memory
            PLDAAPI.PldaMemoryWrite32(DeviceIndex, RegisterOffset.BAR_REGISTER, RegisterOffset.DIO_CONFIGURE / Constants.BUFF_4BYTE_WORDS, Constants.BUFFER_SIZE_LENGTH, WriteBuffer);

        }

        /// <summary>
        /// DIO_Write is used to write the data into DIO registers
        /// </summary>
        /// <param name="DeviceIndex">Index of the device</param>
        /// <param name="Write_Data">Data to be written into DIO registers</param>
        public static void DIO_Write(UInt32 DeviceIndex, int Write_Data)
        {
            UInt32[] WriteBuffer = new UInt32[1];
            UInt32 RegVal;
            UInt32 Wb = 0;

            RegVal = Convert.ToUInt32(Write_Data);
            WriteBuffer[Wb] = RegVal;
            //Writes Data to the memory
            PLDAAPI.PldaMemoryWrite32(DeviceIndex, RegisterOffset.BAR_REGISTER, RegisterOffset.DIO_DATA / Constants.BUFF_4BYTE_WORDS, Constants.BUFFER_SIZE_LENGTH, WriteBuffer);

        }

        /// <summary>
        /// DAC_Update method is called when there is a change in trackbar position
        /// </summary>
        /// <param name="DeviceIndex">Index of device</param>
        /// <param name="DACChannelIndex">Selected channel Index</param>
        /// <param name="Volts">Selected voltage using trackbar</param>
        /// <param name="Index">Channel Index</param>
        public static void DAC_Update(UInt32 DeviceIndex, Int32 DACChannelIndex, double Volts, Int32 Index)
        {
            float Max_Volts = 0;
            float Min_Volts = 0;
            UInt32 Span = 0;
            switch (Index)
            {
                case 0:
                    Max_Volts = 5;
                    Min_Volts = 0;
                    break;
                case 1:
                    Max_Volts = 10;
                    Min_Volts = 0;
                    break;
                case 2:
                    Max_Volts = 5;
                    Min_Volts = -5;
                    break;
                case 3:
                    Max_Volts = 10;
                    Min_Volts = -10;
                    break;
                case 4:
                    Max_Volts = 2.5F;
                    Min_Volts = -2.5F;
                    break;
                default:
                    break;
            }

            Span = (((UInt32)Index) | ((UInt32)DACChannelIndex << DAC_CONFIGURATION.DAC_INDEX_INDICATOR) | (SpanCode.SPAN_TO_N << DAC_CONFIGURATION.SPAN_UPDATE));
            Memory_Write(DeviceIndex, RegisterOffset.WRITE_OFFSET, Span);
            UInt32 res = Convert.ToUInt32(string.Format("{0}", (Constants.STEP_SIZE * ((Volts - Min_Volts) / (Max_Volts - Min_Volts)))));

            UInt32 DACBitCode = ((res) | ((UInt32)DACChannelIndex << DAC_CONFIGURATION.DAC_INDEX_INDICATOR) | (SpanCode.WRITE_UPDATE_N << DAC_CONFIGURATION.SPAN_UPDATE));
            Memory_Write(DeviceIndex, RegisterOffset.WRITE_OFFSET, DACBitCode);

        }
        
        /// <summary>
        /// Close Function is called to close the PLDA device which releases all interrupts,
        /// Clear Resources and writes system reset value to reset the system
        /// </summary>
        /// <param name="DeviceIndex">Index of the Device</param>
        public static void Close(UInt32 DeviceIndex)
        {
            Error_Handler_Close = true;
            UInt32[] CloseWriteBuffer = new UInt32[1];
            UInt32 Wb = 0;

            CloseWriteBuffer[Wb] = Constants.SYSTEMRESET;
            PLDAAPI.PldaMemoryWrite32(DeviceIndex, (UInt32)BAR.BAR_2, RegisterOffset.RESET_REG / Constants.BUFF_4BYTE_WORDS, Constants.BUFFER_SIZE_LENGTH, CloseWriteBuffer);
            PLDAAPI.PldaReleaseInterrupt(DeviceIndex);
            PLDAAPI.PldaClearResources(DeviceIndex);
        }
#if DEBUGFORM
        /// <summary>
        /// Debug_Read function is used to read the registers based on the user's Bar and offset selection
        /// </summary>
        /// <param name="DeviceIndex">Index of the device</param>
        /// <param name="Bar_Reg">Bar Number selected by the user</param>
        /// <param name="Reg_Offset">Reg_Offset selected by the user</param>
        /// <param name="ReadSize">Data Size to be read</param>
        /// <param name="Read_Val">Read value based on the user's Bar and offset selection</param>
        public static void Debug_Read(UInt32 DeviceIndex, UInt32 Bar_Reg, UInt32 Reg_Offset, UInt32 ReadSize, out UInt32[] Read_Val)
        {
            UInt32[] ReadBuffer = new UInt32[128];

            for (UInt32 Byte = 0; Byte < ReadSize; Byte++)
            {
                PLDAAPI.PldaMemoryRead32(DeviceIndex, Bar_Reg, (Reg_Offset / Constants.BUFF_4BYTE_WORDS) + Byte, Constants.BUFFER_SIZE_LENGTH, ref ReadBuffer[Byte]);
            }
            Read_Val = ReadBuffer;
        }

        /// <summary>
        /// Debug_Write function is used to write the registers based on the user's Bar and offset selection
        /// </summary>
        /// <param name="DeviceIndex">Index of the device</param>
        /// <param name="Bar_Reg">Bar Number selected by the user</param>
        /// <param name="Reg_Offset">Reg_Offset selected by the user</param>
        /// <param name="Write_Val">Written value based on the user's Bar and offset selection</param>
        public static void Debug_Write(UInt32 DeviceIndex, UInt32 Bar_reg, UInt32 Reg_Offset, UInt32 Write_Val)
        {
            UInt32[] WriteBuffer = new UInt32[1024];
            UInt32 Wb = 0;

            WriteBuffer[Wb] = Write_Val;

            PLDAAPI.PldaMemoryWrite32(DeviceIndex, Bar_reg, Reg_Offset / Constants.BUFF_4BYTE_WORDS, Constants.BUFFER_SIZE_LENGTH, WriteBuffer);

        }
#endif
    }
}
