unit MainUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, ImgList, Buttons, StdCtrls;

type
  TMainForm = class(TForm)
    LevelImages: TImageList;
    ReadTime: TTimer;
    GroupBox1: TGroupBox;
    DeviceCombo: TComboBox;
    GroupBox2: TGroupBox;
    LevelImage1: TImage;
    LevelImage2: TImage;
    LevelImage3: TImage;
    LevelImage4: TImage;
    LevelImage5: TImage;
    LevelImage6: TImage;
    LevelImage7: TImage;
    LevelImage8: TImage;
    OutButtonA: TSpeedButton;
    BitLabelA7: TLabel;
    BitLabelA6: TLabel;
    BitLabelA5: TLabel;
    BitLabelA4: TLabel;
    BitLabelA3: TLabel;
    BitLabelA2: TLabel;
    BitLabelA1: TLabel;
    BitLabelA0: TLabel;
    GroupBox3: TGroupBox;
    LevelImage9: TImage;
    LevelImage10: TImage;
    LevelImage11: TImage;
    LevelImage12: TImage;
    LevelImage13: TImage;
    LevelImage14: TImage;
    LevelImage15: TImage;
    LevelImage16: TImage;
    OutButtonB0: TSpeedButton;
    OutButtonB1: TSpeedButton;
    OutButtonB2: TSpeedButton;
    BitLabelB7: TLabel;
    BitLabelB6: TLabel;
    BitLabelB5: TLabel;
    BitLabelB4: TLabel;
    BitLabelB3: TLabel;
    BitLabelB2: TLabel;
    BitLabelB1: TLabel;
    BitLabelB0: TLabel;
    OutButtonB3: TSpeedButton;
    OutButtonB4: TSpeedButton;
    OutButtonB5: TSpeedButton;
    OutButtonB6: TSpeedButton;
    OutButtonB7: TSpeedButton;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure OutButtonClick(Sender: TObject);
    procedure ReadTimeTimer(Sender: TObject);
    procedure LevelImageClick(Sender: TObject);
    procedure DeviceComboChange(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  MainForm: TMainForm;

implementation

{$R *.DFM}

uses
  AIOAIO;

const
  DIODataOffset = $44;
  DIOControlOffset = $48;

var
  DIOData: LongWord;
  iBoard: TBoardIndex;
  LevelImage: array[0..15] of TImage;
  OutButton: array[0..8] of TSpeedButton;
  LowLevel, HighLevel: TBitmap;
  bForceUpdate: Boolean;

procedure TMainForm.FormCreate(Sender: TObject);
var
  I: Integer;
  DeviceStr: AnsiString;
  DeviceStrSize: TUInt32;
  Status: TWinErrCode;

begin
  LowLevel := TBitmap.Create;
  HighLevel := TBitmap.Create;
  DeviceCombo.Align := alTop;
  DeviceCombo.Items.Clear;

  LevelImage[ 0] := LevelImage1;
  LevelImage[ 1] := LevelImage2;
  LevelImage[ 2] := LevelImage3;
  LevelImage[ 3] := LevelImage4;
  LevelImage[ 4] := LevelImage5;
  LevelImage[ 5] := LevelImage6;
  LevelImage[ 6] := LevelImage7;
  LevelImage[ 7] := LevelImage8;
  LevelImage[ 8] := LevelImage9;
  LevelImage[ 9] := LevelImage10;
  LevelImage[10] := LevelImage11;
  LevelImage[11] := LevelImage12;
  LevelImage[12] := LevelImage13;
  LevelImage[13] := LevelImage14;
  LevelImage[14] := LevelImage15;
  LevelImage[15] := LevelImage16;
  OutButton[0] := OutButtonA;
  OutButton[1] := OutButtonB0;
  OutButton[2] := OutButtonB1;
  OutButton[3] := OutButtonB2;
  OutButton[4] := OutButtonB3;
  OutButton[5] := OutButtonB4;
  OutButton[6] := OutButtonB5;
  OutButton[7] := OutButtonB6;
  OutButton[8] := OutButtonB7;

  LevelImages.GetBitmap(0, LowLevel);
  LevelImages.GetBitmap(1, HighLevel);
  
  for I := 0 to BRD_GetHighestBoardIndex do begin
    DeviceStrSize := MAX_PATH;
    DeviceStr := StringOfChar(#0, DeviceStrSize);
    Status := BRD_GetName(I, DeviceStrSize, TPAnsiChar(DeviceStr), @DeviceStrSize);
    if Status <> ERROR_SUCCESS then
      DeviceStr := '<Error ' + IntToStr(Status) + ' from BRD_GetName>'
    else
      SetLength(DeviceStr, DeviceStrSize)
    ;
    DeviceCombo.Items.Add('[' + IntToStr(I) + '] ' + DeviceStr);
  end;
  if DeviceCombo.Items.Count <> 0 then begin
    DeviceCombo.ItemIndex := 0;
    DeviceCombo.OnChange(DeviceCombo);
  end;

  bForceUpdate := True;
  ReadTime.Enabled := True;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  LowLevel.Free;
  HighLevel.Free;
end;

procedure TMainForm.OutButtonClick(Sender: TObject);
var
  iOut: Integer;
  DIOControl, Mask: TUInt32;

begin
  //Read back DIO control register, so we can modify it and write it back.
  DIOControl := PIO_In32(iBoard, DIOControlOffset);

  Mask := 1; //Bit 0.
  if OutButton[0].Down then
    DIOControl := DIOControl or Mask
  else
    DIOControl := DIOControl and not Mask
  ;
  for iOut := 1 to 8 do begin
    Mask := 1 shl (7+iOut); //Bits 8..15.
    if OutButton[iOut].Down then
      DIOControl := DIOControl or Mask
    else
      DIOControl := DIOControl and not Mask
    ;
  end;

  //Set new output byte/bit high, in case user made a mistake.
  iOut := (Sender as TSpeedButton).Tag;
  if iOut = 0 then
    DIOData := DIOData or $FF
  else
    DIOData := DIOData or (1 shl (7+iOut))
  ;
  PIO_Out32(iBoard, DIODataOffset, DIOData);

  //Write back DIO control register.
  PIO_Out32(iBoard, DIOControlOffset, DIOControl);

  bForceUpdate := True;
end;

procedure TMainForm.ReadTimeTimer(Sender: TObject);
var
  iDataBit: Integer;
  oDIOData, Changes: LongWord;

begin
  //Get the current DIO data, and a mask of the changed bits to update the display.
  oDIOData := DIOData;
  DIOData := PIO_In32(iBoard, DIODataOffset);
  Changes := DIOData xor oDIOData;

  if bForceUpdate then begin
    Changes := $FFFFFFFF;
    bForceUpdate := False;
  end;

  //Update display for the changed bits.
  for iDataBit := 0 to 15 do
    if (Changes and (1 shl iDataBit)) <> 0 then begin
      if (DIOData and (1 shl iDataBit)) <> 0 then
        LevelImage[iDataBit].Picture.Assign(HighLevel)
      else
        LevelImage[iDataBit].Picture.Assign(LowLevel)
      ;
    end
  ;
end;

procedure TMainForm.LevelImageClick(Sender: TObject);
var
  iDataBit: Integer;

begin
  //Toggle that bit.
  iDataBit := (Sender as TImage).Tag;
  DIOData := DIOData xor (1 shl iDataBit);

  PIO_Out32(iBoard, DIODataOffset, DIOData);

  bForceUpdate := True;
end;

procedure SetOutButtonWithoutOnClick(iOutButton: Integer; bNewState: TUInt32);
var
  BakOnClick: TNotifyEvent;
begin
  BakOnClick := OutButton[iOutButton].OnClick;
  OutButton[iOutButton].OnClick := nil;

  OutButton[iOutButton].Down := bNewState <> 0;

  OutButton[iOutButton].OnClick := BakOnClick;
end;

procedure TMainForm.DeviceComboChange(Sender: TObject);
var
  DIOControl: TUInt32;
  iOutButton: Integer;

begin
  iBoard := DeviceCombo.ItemIndex;

  DIOControl := PIO_In32(iBoard, DIOControlOffset);
  SetOutButtonWithoutOnClick(0, DIOControl and 1);
  for iOutButton := 1 to 8 do
    SetOutButtonWithoutOnClick(iOutButton, DIOControl and (1 shl (7+iOutButton)))
  ;
end;

end.
