unit MainUnit;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Classes, Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs,
  Vcl.StdCtrls, Vcl.ExtCtrls;

type
  TMainForm = class(TForm)
    Panel1: TPanel;
    PollTime: TTimer;
    VoltsList: TListBox;
    Panel2: TPanel;
    GoButton: TButton;
    StopButton: TButton;
    RangeRadios: TRadioGroup;
    GroupBox1: TGroupBox;
    DeviceCombo: TComboBox;
    procedure FormCreate(Sender: TObject);
    procedure GoButtonClick(Sender: TObject);
    procedure StopButtonClick(Sender: TObject);
    procedure PollTimeTimer(Sender: TObject);
  protected
    { Private declarations }
  public
    { Public declarations }
  end;

var
  MainForm: TMainForm;

implementation

{$R *.DFM}

uses
  AIOAIO;

var
  iBoard: TBoardIndex;
  Range: TRangeCode;

procedure TMainForm.FormCreate(Sender: TObject);
var
  I: Integer;
  DeviceStr: AnsiString;
  DeviceStrSize: TUInt32;
  Status: TWinErrCode;

begin
  DeviceCombo.Items.Clear;
  for I := 0 to BRD_GetHighestBoardIndex do begin
    DeviceStrSize := MAX_PATH;
    DeviceStr := StringOfChar(AnsiChar(#0), DeviceStrSize);
    Status := BRD_GetName(I, DeviceStrSize, TPAnsiChar(DeviceStr), @DeviceStrSize);
    if Status <> ERROR_SUCCESS then
      DeviceStr := '<Error ' + AnsiString(IntToStr(Status)) + ' from BRD_GetName>'
    else
      SetLength(DeviceStr, DeviceStrSize)
    ;
    DeviceCombo.Items.Add('[' + IntToStr(I) + '] ' + String(DeviceStr));
  end;
  if DeviceCombo.Items.Count <> 0 then DeviceCombo.ItemIndex := 0;
end;

procedure TMainForm.GoButtonClick(Sender: TObject);
begin
  iBoard := DeviceCombo.ItemIndex;
  case RangeRadios.ItemIndex of
    0: Range := rcBipolar24;
    1: Range := rcBipolar10;
    2: Range := rcBipolar5;
    3: Range := rcBipolar2_5;
    4: Range := rcBipolar1_25;
    5: Range := rcBipolar_625;
    6: Range := rcBipolar20;
  end;

  BRD_Reset(DeviceCombo.ItemIndex);

  PollTime.Enabled := True;
end;

procedure TMainForm.StopButtonClick(Sender: TObject);
begin
  PollTime.Enabled := False;
end;

procedure TMainForm.PollTimeTimer(Sender: TObject);
var
  iCh: TChannelIndex;
  Status: TWinErrCode;
  V: array of TDouble;

begin
  SetLength(V, 8);
  Status := ADC_GetImmediateScanV(iBoard, @V[0], Range);
  if Status <> ERROR_SUCCESS then begin
    VoltsList.Items.Text := 'Error ' + IntToStr(Status) + '.';
    PollTime.Enabled := False;
    Exit;
  end;

  if VoltsList.Items.Count < Length(V) then
    for iCh := VoltsList.Items.Count to High(V) do
      VoltsList.Items.Add('')
  ;
  for iCh := 0 to High(V) do
    VoltsList.Items[iCh] := 'Ch ' + IntToStr(iCh) + ': ' + FloatToStr(V[iCh]) + ' V'
  ;
end;

end.
